#! /usr/bin/python -u
# Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
#
# To activate this script, snmpd must be running and this
# script should be installed /usr/share/snmp/cumulus-sensor.py
# and include the following line in /etc/snmp/snmpd.conf
#
#   pass_persist 1.3.6.1.4.1.40310.6  /usr/share/snmp/cumulus-sensor.py
#

import subprocess
import sys
import json
import time
import types
import os
import syslog
from collections import OrderedDict
import getopt

import snmp_passpersist as snmp

short_options = "di:a"
long_options = ["debug", "interval="]
refreshInterval = 20

# this is the Entity Sensor MIB (RFC 3433) which is mostly the output
# of the smonctl command
entitySensorMIB = '.1.3.6.1.4.1.40310.6'

sensorIDStart = {'temp' :100000000,
                 'fan'  :100011000,
                 'power':110000000,
                 'volt' :110001000}

intermediateValueDict = {}
# some typical types
celsius = 8
watts = 6
rpm = 10
voltsAC = 3
voltsDC = 4

# just the value, no multipliers
units = 9

# precision, meaning the number of decimal places in the
# fractional part of an associated EntitySensorValue fixed-
# point number
precision_zero = 0
precision_one = 1
precision_two = 2

# status is either ok, etc...
status_ok = 1
status_unavailable = 2
status_nonoperational = 3

# Admin status
adminStatusNa = 2
adminStatusOn = 1
adminStatusOff = 0

status = {'OK':status_ok,
          'ABSENT':status_unavailable,
          'BAD':status_nonoperational}

#alarm values
other = 1
unknown = 2
normal = 3
warning = 4
alert = 5
critical = 6
NotPresent = 7
NotOperational = 8
unavailable = 9

valueArray = {'index':{'temp':None,
                       'fan':None,
                       'power':None,
                       'volt':None,
                       'oid_type':'add_int'},
              'datatype':{'temp':celsius,
                          'fan':rpm,
                          'power':watts,
                          'volt':voltsDC,
                          'oid_type':'add_int'},
              'scale':{'temp':units,
                       'fan':units,
                       'power':units,
                       'volt':units,
                       'oid_type':'add_int'},
              'precision':{'temp':precision_one,
                           'fan':precision_zero,
                           'power':precision_two,
                           'volt':precision_two,
                           'oid_type':'add_int'},
              'value':{'temp':'input',
                       'fan':'input',
                       'power':'input',
                       'volt':'input',
                       'oid_type':'add_int'},
              'operational_status':{'temp':'state',
                                    'fan':'state',
                                    'power':'state',
                                    'volt':'state',
                                    'oid_type':'add_int'},
              'units_display':{'temp':'Celsius',
                               'fan':'RPM',
                               'power':'Watts',
                               'volt':'Volts',
                               'oid_type':'add_str'},
              'value_timestamp':{'temp':None,
                                 'fan':None,
                                 'power':None,
                                 'volt':None,
                                 'oid_type':'add_tt'},
              'update_rate':{'temp':'15000',
                             'fan':'15000',
                             'power':'15000',
                             'volt':'15000',
                             'oid_type':'add_gau'},
              'description':{'temp':'description',
                       'fan':'description',
                       'power':'description',
                       'volt':'description',
                       'oid_type':'add_str'},
              'max':{'temp':'max',
                       'fan':'max',
                       'power':'max',
                       'volt':'max',
                       'oid_type':'add_int'},
              'min':{'temp':'min',
                       'fan':'min',
                       'power':'min',
                       'volt':'min',
                       'oid_type':'add_int'},
              'alarm':{'temp':None,
                       'fan':None,
                       'power':None,
                       'volt':None,
                       'oid_type':'add_int'},
              'crit':{'temp':'crit',
                      'fan':'crit',
                      'power':'crit',
                      'volt':'crit',
                      'oid_type':'add_int'},
              'adminStatus':{'temp':'state',
                      'fan':'state',
                      'power':'state',
                      'volt':'state',
                      'oid_type':'add_int'},
              'lcrit':{'temp':'lcrit',
                      'fan':'lcrit',
                      'power':'lcrit',
                      'volt':'lcrit',
                      'oid_type':'add_int'},
              }
# Alarm level calculation function
# normal level : value is in between min and max thresholds
# Alert : sensor value is in between max and crit or min and lcrit
# Critical: Sensor value is above crit level or below lcrit level

def getAlarmLevel(rowname, sensorType, sensor, uptime, sensorID):
    alarm = 3
    crit_val = 0
    critValValid = False
    lcrit_val = 0
    lcritValValid = False

    temp_dict = intermediateValueDict.get(sensorID, None)
    if temp_dict == None:
        return 1

    min_val = temp_dict.get('min')
    max_val = temp_dict.get('max')
    act_val = temp_dict.get('value')

    status_val = status.get(sensor.get(valueArray['operational_status'].get(sensorType, None), 0),
                                 status_unavailable)
    if status_val != status_ok:
        return unavailable 

    multiplier = 10 ** int(valueArray['precision'][sensorType])
    inputval = sensor.get(valueArray['crit'].get(sensorType, None), 0)

    if inputval and (isinstance(inputval, types.IntType) or \
                     isinstance(inputval, types.FloatType)):
        crit_val = int(inputval * multiplier)
        critValValid = True
    else:
        crit_val = 0
        critValValid = False

    multiplier = 10 ** int(valueArray['precision'][sensorType])
    inputval = sensor.get(valueArray['lcrit'].get(sensorType, None), 0)

    if inputval and (isinstance(inputval, types.IntType) or \
                     isinstance(inputval, types.FloatType)):
        lcrit_val = int(inputval * multiplier)
        lcritValValid = True
    else:
        lcrit_val = 0
        lcritValValid = False

    if min_val == 0 and max_val == 0 and act_val == 0:
        alarm = unknown
        return alarm

    if critValValid == True:
        if act_val >= crit_val:
            alarm = critical
            return alarm
        elif crit_val > act_val >= max_val:
            alarm = alert
            return alarm

    if lcritValValid == True:
        if act_val <= lcrit_val:
            alarm = critical
            return alarm
        elif lcrit_val < act_val <= min_val:
            alarm = alert
            return alarm

    if min_val < act_val < max_val:
        alarm = normal
    else:
        alarm = critical
    return alarm


def getValue(rowname, sensorType, sensor, uptime, sensorID):
    if rowname == 'value' or rowname == 'max' or rowname == 'min':
        # this is the only case where we return an actual number, or 0
        # we need to make sure this is an integer value and scaled by
        # the precision
        multiplier = 10 ** int(valueArray['precision'][sensorType])
        inputval = sensor.get(valueArray[rowname].get(sensorType, None), 0)
        if inputval and (isinstance(inputval, types.IntType) or \
                         isinstance(inputval, types.FloatType)):
            val = int(inputval * multiplier)
        else:
            val = 0
        # Creating a local dict to store min max and the value of each sensor
        temp_dict = {rowname:val}
        temp_outer_dict = intermediateValueDict.get(sensorID,None)
        if (temp_outer_dict == None):
            temp_outer_dict = {}
        else:
            temp_outer_dict = intermediateValueDict.get(sensorID,None)
        temp_outer_dict.update(temp_dict)
        temp_outer_dict = {sensorID:temp_outer_dict}
        intermediateValueDict.update(temp_outer_dict)

    elif rowname == 'operational_status':
        # here, we return the numerical value, we map (via status) to OK=1, etc.
        # if status does not exist, we see it as unavailable
        val = status.get(sensor.get(valueArray[rowname].get(sensorType, None), 0),
                         status_unavailable)
    elif rowname == 'adminStatus':
        #By default we are setting it to on as we 
        #do not have any admin status for PSUs/Fans
        if sensorType == 'temp':
            val = adminStatusNa
        else:
            val = adminStatusOn
    elif rowname == 'description':
        val = sensor.get(valueArray[rowname].get(sensorType, None), 0)
    elif rowname == 'value_timestamp':
        val = uptime
    elif rowname == 'alarm':
        val = getAlarmLevel(rowname, sensorType, sensor, uptime, sensorID)
    elif rowname == 'index':
        val = sensorID
    else:
        # these are mostly statics
        # we could get a new sensor type so we need to default without an error
        val = valueArray[rowname].get(sensorType,0)
    return(val)

def update():
    """
    Simply grab the output of smonctl -j and stick it in our hashed array
    """
    oid = '1.1.1'
    # these are the elements of the table we will loop over
    entPhySensorTable = enumerate(['index', 'datatype', 'scale', 'precision', 'value',
                                   'operational_status', 'units_display',
                                   'value_timestamp', 'update_rate', 'description','min','max','alarm','adminStatus'],
                                  start=1)

    # this is just a list of dictionaries, one for each sensor
    try:
        smonctlJsonList = json.loads(subprocess.check_output((['/usr/sbin/smonctl', '-j']),
                                                         stderr=subprocess.STDOUT,
                                                         shell=False))
    except Exception as e:
        syslog.syslog("cumulus-sensor.py: Exception=%s" % e)
        smonctlJsonList = []

    hwTypes = OrderedDict([('temp', []), ('fan', []), ('power', []), ('volt', [])])
    # reorganize the data by sensor type in hwTypes dict
    # and give each sensor an ID
    allTypes = hwTypes.keys()
    for sensor in smonctlJsonList:
        # create the index if needed
        # for now, we only keep the types we know about
        if sensor['type'] not in allTypes:
            # we do not know about this type, so skip it.
            syslog.syslog("cumulus-sensor.py: Missing support for type=%s" % sensor['type'])
            continue
        hwTypes.setdefault(sensor['type'],[]).append(sensor)

    # grab the uptime timetics since smonctl -j does not actually set it
    try:
        fd = open('/proc/uptime', 'r')
        uptime = int(float(fd.read().split()[0]) * 100)
        fd.close()
    except Exception as e:
        syslog.syslog("cumulus-sensor.py: Exception=%s" % e)
        uptime = 0

    for rowval,rowname in entPhySensorTable:  # 1, units
        for sensorType in allTypes:     # temp, fan, power, volt
            # set the sensor ID, this is order dependent based on hwTypes
            sensorID = sensorIDStart[sensorType]
            for sensor in hwTypes[sensorType]:# actual sensor arrays
                # make sure to update the sensorID
                sensorID += 1
                myval = getValue(rowname, sensorType, sensor, uptime, sensorID)
                # call the method defined in oid_type
                myfunc = getattr(pp, valueArray[rowname]['oid_type'])
                # here, we will call add_int, add_str, add_tt, or add_gau
                myfunc('%s.%s.%s' % (oid, rowval, sensorID), myval)

syslog.syslog("cumulus-sensor.py: starting...")
pp = snmp.PassPersist(entitySensorMIB)
pp.debug = False

full_cmd_arguments = sys.argv
argument_list = full_cmd_arguments[1:]

try:
    arguments, values = getopt.getopt(argument_list, short_options, long_options)
except getopt.error as err:
    syslog.syslog(str(err))
    sys.exit(2)

for current_argument, current_value in arguments:
    if current_argument in ("-d", "--debug"):
        pp.debug = True
    elif current_argument in ("-i", "--interval"):
        refreshInterval = float(current_value)

if pp.debug:
    update()
else:
    syslog.syslog("cumulus-sensor.py: initialized...")
    pp.start(update, refreshInterval)
    syslog.syslog("cumulus-sensor.py: after first update, we should not see this...")

