#! /usr/bin/python -u
# Copyright (C) 2020-2021 NVIDIA Corporation. ALL RIGHTS RESERVED.
#
# To activate this script, snmpd must be running and this
# script should be installed /usr/share/snmp/cumulus-status.py
# and include the following line in /etc/snmp/snmpd.conf

import subprocess
import sys
import json
import time
import types
import os
import syslog
from collections import OrderedDict
import getopt
from queue import Queue
import traceback

import snmp_passpersist as snmp

short_options = "di:a"
long_options = ["debug", "interval="]
refreshInterval = 5

utzInterval = 60
# this is the Entity Sensor MIB (RFC 3433) which is mostly the output
# of the smonctl command
entitySensorMIB = '.1.3.6.1.4.1.40310.5'
errorCodes = {
        "other": 1,
        "unknown": 2,
        "normal": 3,
        "warning": 4,
        "alert": 5,
        "critical": 6,
        "NotPresent": 7,
        "NotOperational": 8
}

memPrecision = 0
cpuPrecision = 2

cpuDict = { 0:'user',
            1:'nice',
            2:'system',
            3:'idle',
            4:'iowait',
            5:'irq',
            6:'softirq',
            7:'steal',
            8:'guest',
            9:'guest_nice'}

utilQ = Queue(maxsize = 60)

dataDictOld = {}
dataDictNew = {}

def getProcStat():
    dataDict = dict()
    try:
        with open('/proc/stat', 'r') as procfile:
            cputimes = procfile.readline()
            j = 0
            for i in cputimes.split(' ')[2:]:
                key = cpuDict[j]
                dataDict[key] = int(i)
                j = j + 1
            return dataDict
    except Exception as e:
        syslog.syslog('ERROR: %s' % e)
        syslog.syslog(traceback.format_exc())
        sys.exit(3)

def calcCPUUtil():
    try:
        global dataDictOld
        global dataDictNew
        global utilQ

        dataDictNew = getProcStat()

        idleOld = dataDictOld['idle']
        idleNew = dataDictNew['idle']

        nonIdleOld = dataDictOld['user'] + dataDictOld['nice'] + \
            dataDictOld['system'] + dataDictOld['irq']  + \
            dataDictOld['softirq'] + dataDictOld['steal'] + \
            dataDictNew['iowait']
        nonIdleNew = dataDictNew['user'] + dataDictNew['nice'] + \
            dataDictNew['system'] + dataDictNew['irq']  + \
            dataDictNew['softirq'] + dataDictNew['steal'] + \
            dataDictNew['iowait']

        totalOld = idleOld + nonIdleOld
        totalNew = idleNew + nonIdleNew

        totald = totalNew - totalOld
        idled = idleNew - idleOld

        if totald:
            cpu_use = int (100.0 * ((float(totald) - float(idled))/ float(totald)))
        else:
            cpu_use = 0
        dataDictOld = dataDictNew

        if utilQ.full():
            utilQ.get()

        utilQ.put(cpu_use)
        return cpu_use
    except Exception as e:
        syslog.syslog('ERROR: %s' % (e))
        syslog.syslog(traceback.format_exc())
        return 0

def calcOneMinUtil():
    try:
        global utilQ
        hist = list(utilQ.queue)
        hist.reverse()
        i = 0
        total = 0
        qlen = len(hist)
        if qlen > 12:
            qlen = 12
        while i < qlen:
            total = total + hist[i]
            i = i + 1
        return int(total/qlen)
    except Exception as e:
        syslog.syslog('ERROR: %s' % (e))
        syslog.syslog(traceback.format_exc())
        return 0

def calcFiveMinUtil():
    try:
        global utilQ
        hist = list(utilQ.queue)
        hist.reverse()
        i = 0
        total = 0
        qlen = len(hist)
        if qlen > 60:
            qlen = 60
        while i < qlen:
            total = total + hist[i]
            i = i + 1
        return int(total/qlen)
    except Exception as e:
        syslog.syslog('ERROR: %s' % (e))
        syslog.syslog(traceback.format_exc())
        return 0

def getCPUCores():
    with open('/proc/cpuinfo') as file:
        for line in file:
            if 'cpu core' in line:
                numCore = line.split()[3]
                break
    return int(numCore) 


def getCpuStatus(oneMinUtz):
    fullUtz = 100

    if oneMinUtz <= fullUtz/4:
        return errorCodes["normal"]
    elif fullUtz/4 <= oneMinUtz < fullUtz/2:
        return errorCodes["warning"]
    elif fullUtz/2 <= oneMinUtz < fullUtz*3/4:
        return errorCodes["alert"]
    elif fullUtz*3/4 <= oneMinUtz :
        return errorCodes["critical"]

def getMemStatus(totalMem, freeMem, availMem):
    freeMemPercentage = int (int(freeMem)*100/int(totalMem))
    if freeMemPercentage >= 75:
        return errorCodes["normal"]
    elif 75 > freeMemPercentage >= 50:
        return errorCodes["warning"]
    elif 50 > freeMemPercentage >= 25:
        return errorCodes["alert"]
    elif 25 > freeMemPercentage:
        return errorCodes["critical"]

def checkMemThreshold(totalMem, freeMem, threshMem):
    
    if int(freeMem) < int(threshMem):
        return 1
    else:
        return 0

def checkCPUThreshold(risingThreshold,oneMinUtz):
    if oneMinUtz > risingThreshold:
        return 1
    else:
        return 0

def update():

    fiveSecUtil = calcCPUUtil()
    oneMinUtil = calcOneMinUtil()
    fiveMinUtil =  calcFiveMinUtil()

    loadAvgStr = str(fiveSecUtil) + ", " + str(oneMinUtil) + " ," + str(fiveMinUtil)

    oid = '1.1'
    notification_oid = '2'
    agentSwitchCpuProcessMemFree = '1'
    agentSwitchCpuProcessMemAvailable = '2'
    agentSwitchCpuProcessMemTotal = '3'
    agentSwitchCpuProcessMemPrecision = '4'
    agentSwitchCpuProcessRisingThreshold = '5'
    agentSwitchCpuProcessRisingThresholdInterval = '6'
    agentSwitchCpuProcessFallingThreshold = '7'
    agentSwitchCpuProcessFallingThresholdInterval = '8'
    agentSwitchCpuProcessFreeMemoryThreshold = '9'
    agentSwitchCpuProcessTotalUtilization = '10'

    agentSwitchCpuProcess5SecUtilization = '11'
    agentSwitchCpuProcess1minUtilization = '12'
    agentSwitchCpuProcess5minUtilization = '13'
    agentSwitchCpuProcess15minUtilization = '14'
    agentSwitchCpuUtzPrecision = '15'

    agentSwitchCpuCores = '16'

    agentSwitchCPUUtzErrorFlag = '100'
    agentSwitchCPUUtzErrorMsg = '101'
    agentSwitchMemErrorFlag = '102'
    agentSwitchMemErrorMsg = '103'

    agentSwitchCpuutilizationStatus = '1'
    agentSwitchMemStatus = '2'

    with open('/proc/meminfo') as file:
       for line in file:
           if 'MemFree' in line:
               free_mem_in_kb = line.split()[1]
           elif 'MemTotal' in line:
               total_mem_in_kb = line.split()[1]
           elif 'MemAvailable' in line:
               avail_mem_in_kb = line.split()[1]

    coreCnt = getCPUCores()
    risingThreshold = 75
    fallingThreshold = 75

    memThreshold = int(float(total_mem_in_kb) * 0.4)

    cpuUtzStatus = getCpuStatus(oneMinUtil)
    memUtzStatus = getMemStatus(total_mem_in_kb,free_mem_in_kb,avail_mem_in_kb)

    myfunc = getattr(pp, 'add_int')
    myfunc('%s.%s' % (oid, agentSwitchCpuProcessMemFree),free_mem_in_kb)
    myfunc('%s.%s' % (oid, agentSwitchCpuProcessMemAvailable),avail_mem_in_kb)
    myfunc('%s.%s' % (oid, agentSwitchCpuProcessMemTotal), total_mem_in_kb)
    myfunc('%s.%s' % (oid, agentSwitchCpuProcessMemPrecision), memPrecision)

    myfunc = getattr(pp, 'add_gau')
    myfunc('%s.%s' % (oid, agentSwitchCpuProcessRisingThreshold),risingThreshold)
    myfunc('%s.%s' % (oid, agentSwitchCpuProcessFreeMemoryThreshold),memThreshold)
    myfunc('%s.%s' % (oid, agentSwitchCpuProcessFallingThreshold),fallingThreshold)
    myfunc('%s.%s' % (oid, agentSwitchCpuProcessRisingThresholdInterval), utzInterval)
    myfunc('%s.%s' % (oid, agentSwitchCpuProcessFallingThresholdInterval), utzInterval)

    myfunc = getattr(pp, 'add_int')
    myfunc('%s.%s' % (oid, agentSwitchCpuProcess5SecUtilization),fiveSecUtil)
    myfunc('%s.%s' % (oid, agentSwitchCpuProcess1minUtilization),oneMinUtil)
    myfunc('%s.%s' % (oid, agentSwitchCpuProcess5minUtilization),fiveMinUtil)
    myfunc('%s.%s' % (oid, agentSwitchCpuUtzPrecision), cpuPrecision)
    myfunc('%s.%s' % (oid, agentSwitchCpuCores), coreCnt)

    myfunc = getattr(pp, 'add_str')
    myfunc('%s.%s' % (oid, agentSwitchCpuProcessTotalUtilization),loadAvgStr)

    myfunc = getattr(pp, 'add_int')
    myfunc('%s.%s' % (notification_oid, agentSwitchCpuutilizationStatus), cpuUtzStatus)
    myfunc('%s.%s' % (notification_oid, agentSwitchMemStatus), memUtzStatus)


    if checkCPUThreshold(risingThreshold,oneMinUtil):
        myfunc = getattr(pp, 'add_int')
        myfunc('%s.%s' % (oid, agentSwitchCPUUtzErrorFlag), 1)
        myfunc = getattr(pp, 'add_str')
        myfunc('%s.%s' % (oid, agentSwitchCPUUtzErrorMsg), "CPU Utilization exceeds threshold")
    else:
        myfunc = getattr(pp, 'add_int')
        myfunc('%s.%s' % (oid, agentSwitchCPUUtzErrorFlag), 0)
        myfunc = getattr(pp, 'add_str')
        myfunc('%s.%s' % (oid, agentSwitchCPUUtzErrorMsg), "No Error")

    if checkMemThreshold(total_mem_in_kb,free_mem_in_kb,memThreshold):
        myfunc = getattr(pp, 'add_int')
        myfunc('%s.%s' % (oid, agentSwitchMemErrorFlag), 1)
        myfunc = getattr(pp, 'add_str')
        myfunc('%s.%s' % (oid, agentSwitchMemErrorMsg), "Free Memory gone below threshold")
    else:
        myfunc = getattr(pp, 'add_int')
        myfunc('%s.%s' % (oid, agentSwitchMemErrorFlag), 0)
        myfunc = getattr(pp, 'add_str')
        myfunc('%s.%s' % (oid, agentSwitchMemErrorMsg), "No Error")

syslog.syslog("cumulus-status.py: starting...")
pp = snmp.PassPersist(entitySensorMIB)
pp.debug = False

full_cmd_arguments = sys.argv
argument_list = full_cmd_arguments[1:]

try:
    arguments, values = getopt.getopt(argument_list, short_options, long_options)
except getopt.error as err:
    syslog.syslog(str(err))
    sys.exit(2)

for current_argument, current_value in arguments:
    if current_argument in ("-d", "--debug"):
        pp.debug = True
    elif current_argument in ("-i", "--interval"):
        refreshInterval = float(current_value)

dataDictOld = getProcStat()
dataDictNew = getProcStat()

if pp.debug:
    update()
else:
    syslog.syslog("cumulus-status.py: initialized...")
    pp.start(update, refreshInterval)
    syslog.syslog("cumulus-status.py: after first update, we should not see this...")

